/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: TaxonomyFactory.java,v 3cadcf8b23fe 2010/06/11 12:26:37 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.taxonomy;

import java.util.Map.Entry;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.text.Collator;
import java.io.BufferedReader;
import java.io.Reader;
import org.json.me.JSONArray;
import org.json.me.JSONException;
import org.json.me.JSONObject;
import it.tidalwave.util.Id;
import it.tidalwave.bluebill.taxonomy.Taxonomy.Type;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public class TaxonomyFactory
  {
    private final static Map<Type, String> NAME_MAP_BY_TYPE = new HashMap<Type, String>();

    private final static List<Locale> LOCALES = new ArrayList<Locale>();

    private final Map<String, Map<Id, String>> nameBagsByLanguage = new HashMap<String, Map<Id, String>>();

    static
      {
        NAME_MAP_BY_TYPE.put(Type.CLASS, "Class");
        NAME_MAP_BY_TYPE.put(Type.ORDER, "Order");
        NAME_MAP_BY_TYPE.put(Type.FAMILY, "Family");
        NAME_MAP_BY_TYPE.put(Type.GENUS, "Genus");
        NAME_MAP_BY_TYPE.put(Type.SPECIES, "Species");
        NAME_MAP_BY_TYPE.put(Type.SUBSPECIES, "Subspecies");

        final String[] languageCodes = { "en", "en_US", "it", "de", "da", "no", "fi", "sv", "fr", "es", "nl" };

        for (final String languageCode : languageCodes)
          {
            LOCALES.add(new Locale(languageCode));
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final Comparator<Taxon> taxonComparator = new Comparator<Taxon>()
      {
        final Collator collator = Collator.getInstance();

        public int compare (final @Nonnull Taxon taxon1, final @Nonnull Taxon taxon2)
          {
            return collator.compare(taxon1.getScientificName(), taxon2.getScientificName());
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Taxonomy loadTaxonomy (final @Nonnull Reader r)
      throws Exception
      {
        final BufferedReader br = new BufferedReader(r);
        final StringBuilder builder = new StringBuilder();

        for (;;)
          {
            final String line = br.readLine();

            if (line == null)
              {
                break;
              }

            builder.append(line);
          }

        final String string = builder.toString();
        final JSONObject rootObject = new JSONObject(string);
        final JSONObject jsonTaxonomy = rootObject.getJSONObject("Taxonomy");
        final Taxonomy taxonomy = new Taxonomy(jsonTaxonomy.getString("name"));
        scan(taxonomy, null, jsonTaxonomy, 0);

        final Set<Type> types = new HashSet<Type>(taxonomy.taxonListMapByType.keySet());
        types.remove(Type.SPECIES);
        types.remove(Type.SUBSPECIES);

        for (final Type type : types)
          {
            final List<Taxon> taxa = taxonomy.taxonListMapByType.get(type);
            Collections.sort(taxa, taxonComparator);
          }

        for (final Entry<String, Map<Id, String>> entry : nameBagsByLanguage.entrySet())
          {
            final Locale locale = new Locale(entry.getKey());
            taxonomy.namesBagByLocale.put(locale, new LocalizedNamesBag(locale, entry.getValue()));
          }

        nameBagsByLanguage.clear();

        return taxonomy;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Taxon> scan (final @Nonnull Taxonomy taxonomy,
                              final @Nonnull Taxon parent,
                              final @Nullable JSONObject jsonParent,
                              final @Nonnull int level)
      throws JSONException
      {
        final Type type = Type.values()[level];
        final List<Taxon> children = new ArrayList<Taxon>();
        final String childName = NAME_MAP_BY_TYPE.get(type);
        final JSONArray jsonChildren = jsonParent.optJSONArray(childName);

        if (jsonChildren != null)
          {
            for (int i = 0; i < jsonChildren.length(); i++)
              {
                final JSONObject jsonChild = jsonChildren.getJSONObject(i);
                // FIXME: id should be mandatory, but unfortunately Aves hasn't in EBNItalia 2003
                final Id id = new Id(jsonChild.optString("id"));
                final Taxon child = new Taxon(taxonomy, id, jsonChild.getString("name"), type, parent);

                for (final Locale locale : LOCALES)
                  {
                    addDisplayName(jsonChild, child, locale);
                  }

                children.add(child);
                taxonomy.taxonMapById.put(id, child);

                if (level < Type.values().length - 1)
                  {
                    child.children.addAll(scan(taxonomy, child, jsonChild, level + 1));
                  }
              }
          }
        
        List<Taxon> taxa = taxonomy.taxonListMapByType.get(type);
        
        if (taxa == null)
          {
            taxa = children;
          }
        else
          {
            taxa.addAll(children);
          }

        taxonomy.taxonListMapByType.put(type, taxa);

        return children;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void addDisplayName (final @Nonnull JSONObject jsonObject,
                                 final @Nonnull Taxon taxon,
                                 final @Nonnull Locale locale)
      {
        final String displayName = jsonObject.optString("lang_" + locale.getLanguage().toLowerCase());

        if ((displayName != null) && !displayName.equals(""))
          {
            final String language = locale.getLanguage();
            Map<Id, String> nameBag = nameBagsByLanguage.get(language);

            if (nameBag == null)
              {
                nameBag = new HashMap<Id, String>();
                nameBagsByLanguage.put(language, nameBag);
              }

            nameBag.put(taxon.getId(), displayName);
          }
      }
  }
